@@ -5,7 +5,7 @@
 #
 # Copyright (c) 2008-2009 Yu Yuwei <acevery@gmail.com>
 # Copyright (c) 2009-2014 Caius "kaio" CHANCE <me@kaio.net>
-# Copyright (c) 2012-2015 Mike FABIAN <mfabian@redhat.com>
+# Copyright (c) 2012-2022 Mike FABIAN <mfabian@redhat.com>
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -27,6 +27,8 @@ from typing import Any
 from typing import List
 from typing import Tuple
 from typing import Dict
+from typing import Callable
+from enum import Enum, Flag
 import sys
 import os
 import logging
@@ -35,12 +37,11 @@ from gi import require_version # type: ignore
 require_version('Gio', '2.0')
 from gi.repository import Gio # type: ignore
 require_version('GLib', '2.0')
-from gi.repository import GLib # type: ignore
+from gi.repository import GLib
 require_version('Gdk', '3.0')
 from gi.repository import Gdk
 require_version('Gtk', '3.0')
 from gi.repository import Gtk
-from gi.repository import GObject
 require_version('IBus', '1.0')
 from gi.repository import IBus
 
@@ -50,8 +51,8 @@ import tabsqlitedb
 LOGGER = logging.getLogger('ibus-table')
 
 DOMAINNAME = 'ibus-table'
-_ = lambda a: gettext.dgettext(DOMAINNAME, a)
-N_ = lambda a: a
+_: Callable[[str], str] = lambda a: gettext.dgettext(DOMAINNAME, a)
+N_: Callable[[str], str] = lambda a: a
 
 # When matching keybindings, only the bits in the following mask are
 # considered for key.state:
@@ -59,6 +60,7 @@ KEYBINDING_STATE_MASK = (
     IBus.ModifierType.MODIFIER_MASK
     & ~IBus.ModifierType.LOCK_MASK # Caps Lock
     & ~IBus.ModifierType.MOD2_MASK # Num Lock
+    & ~IBus.ModifierType.MOD3_MASK # Scroll Lock
 )
 
 def variant_to_value(variant: GLib.Variant) -> Any:
@@ -191,6 +193,7 @@ def get_default_chinese_mode(database: tabsqlitedb.TabSqliteDb) -> int:
 def get_default_keybindings(
         gsettings: Gio.Settings,
         database: tabsqlitedb.TabSqliteDb) -> Dict[str, List[str]]:
+    default_keybindings: Dict[str, List[str]] = {}
     default_keybindings = variant_to_value(
         gsettings.get_default_value('keybindings'))
     # Now update the default keybindings from gsettings with
@@ -306,7 +309,7 @@ def get_default_keybindings(
     return default_keybindings
 
 def dict_update_existing_keys(
-        pdict: Dict[Any, Any], other_pdict: Dict[Any, Any]):
+        pdict: Dict[Any, Any], other_pdict: Dict[Any, Any]) -> None:
     '''Update values of existing keys in a Python dict from another Python dict
 
     Using pdict.update(other_pdict) would add keys and values from other_pdict
@@ -333,6 +336,317 @@ def dict_update_existing_keys(
         if key in pdict:
             pdict[key] = other_pdict[key]
 
+class Capabilite(Flag):
+    '''Compatibility class to handle IBus.Capabilite the same way no matter
+    what version of ibus is used.
+
+    For example, older versions of ibus might not have IBus.Capabilite.SYNC_PROCESS_KEY
+    (or maybe even do not have IBus.Capabilite at all). Then
+
+        capabilities & IBus.Capabilite.SYNC_PROCESS_KEY
+
+    will produce an exception. But when using this compatibility class
+
+        capabilities & IBus.Capabilite.SYNC_PROCESS_KEY
+
+    will just be False but not cause an exception.
+
+    >>> int(Capabilite.PREEDIT_TEXT)
+    1
+
+    >>> Capabilite.PREEDIT_TEXT == 1
+    True
+
+    >>> Capabilite.PREEDIT_TEXT | 2
+    3
+
+    >>> 2 | Capabilite.PREEDIT_TEXT
+    3
+
+    >>> int(Capabilite.PREEDIT_TEXT | Capabilite.AUXILIARY_TEXT)
+    3
+
+    >>> 3 == Capabilite.AUXILIARY_TEXT | Capabilite.PREEDIT_TEXT
+    True
+
+    >>> 3 == Capabilite.AUXILIARY_TEXT | IBus.Capabilite.PREEDIT_TEXT
+    True
+
+    >>> Capabilite.PREEDIT_TEXT == IBus.Capabilite.PREEDIT_TEXT
+    True
+    '''
+    def __new__(cls, attr: str) -> Any:
+        obj = object.__new__(cls)
+        if hasattr(IBus, 'Capabilite') and hasattr(IBus.Capabilite, attr):
+            obj._value_ = int(getattr(IBus.Capabilite, attr))
+        else:
+            obj._value_ = 0
+        return obj
+
+    def __int__(self) -> int:
+        return int(self._value_)
+
+    def __eq__(self, other: Any) -> bool:
+        if (self.__class__ is other.__class__
+            or other.__class__ is IBus.Capabilite):
+            return bool(int(self) == int(other))
+        if other.__class__ is int or other.__class__ is float:
+            return bool(int(self) == other)
+        return NotImplemented
+
+    def __or__(self, other: Any) -> Any:
+        if self.__class__ is other.__class__:
+            return self.value | other.value
+        if (other.__class__ is IBus.Capabilite):
+            return int(self) | int(other)
+        if other.__class__ is int:
+            return int(self) | other
+        return NotImplemented
+
+    def __ror__(self, other: Any) -> Any: # type: ignore[override]
+        return self.__or__(other)
+
+    def __and__(self, other: Any) -> Any:
+        if self.__class__ is other.__class__:
+            return self.value & other.value
+        if (other.__class__ is IBus.Capabilite):
+            return int(self) & int(other)
+        if other.__class__ is int:
+            return int(self) & other
+        return NotImplemented
+
+    def __rand__(self, other: Any) -> Any: # type: ignore[override]
+        return self.__and__(other)
+
+    PREEDIT_TEXT = ('PREEDIT_TEXT')
+    AUXILIARY_TEXT = ('AUXILIARY_TEXT')
+    LOOKUP_TABLE = ('LOOKUP_TABLE')
+    FOCUS = ('FOCUS')
+    PROPERTY = ('PROPERTY')
+    SURROUNDING_TEXT = ('SURROUNDING_TEXT')
+    OSK = ('OSK')
+    SYNC_PROCESS_KEY = ('SYNC_PROCESS_KEY')
+
+class InputPurpose(Enum):
+    '''Compatibility class to handle InputPurpose the same way no matter
+    what version of ibus is used.
+
+    For example, older versions of ibus might not have
+    IBus.InputPurpose.TERMINAL and then
+
+        input_purpose == IBus.InputPurpose.TERMINAL
+
+    will produce an exception. But when using this compatibility class
+
+        input_purpose == InputPurpose.TERMINAL
+
+    will just be False but not cause an exception.
+
+    See also:
+
+    https://docs.gtk.org/gtk3/enum.InputPurpose.html
+    https://docs.gtk.org/gtk4/enum.InputPurpose.html
+
+    Examples:
+
+    >>> int(InputPurpose.PASSWORD)
+    8
+
+    >>> 8 == InputPurpose.PASSWORD
+    True
+
+    >>> int(InputPurpose.PIN)
+    9
+
+    >>> InputPurpose.PASSWORD <= InputPurpose.PIN
+    True
+
+    >>> InputPurpose.PASSWORD == Gtk.InputPurpose.PASSWORD
+    True
+
+    >>> InputPurpose.PASSWORD == IBus.InputPurpose.PASSWORD
+    True
+    '''
+    def __new__(cls, attr: str) -> Any:
+        obj = object.__new__(cls)
+        if hasattr(Gtk, 'InputPurpose') and hasattr(Gtk.InputPurpose, attr):
+            obj._value_ = int(getattr(Gtk.InputPurpose, attr))
+        else:
+            obj._value_ = -1
+        return obj
+
+    def __int__(self) -> int:
+        return int(self._value_)
+
+    def __eq__(self, other: Any) -> bool:
+        if (self.__class__ is other.__class__
+            or other.__class__ is Gtk.InputPurpose
+            or other.__class__ is IBus.InputPurpose):
+            return int(self) == int(other)
+        if other.__class__ is int or other.__class__ is float:
+            return bool(int(self) == other)
+        return NotImplemented
+
+    def __gt__(self, other: Any) -> bool:
+        if (self.__class__ is other.__class__
+            or other.__class__ is Gtk.InputPurpose
+            or other.__class__ is IBus.InputPurpose):
+            return int(self) > int(other)
+        if other.__class__ is int or other.__class__ is float:
+            return bool(int(self) > other)
+        return NotImplemented
+
+    def __lt__(self, other: Any) -> bool:
+        if (self.__class__ is other.__class__
+            or other.__class__ is Gtk.InputPurpose
+            or other.__class__ is IBus.InputPurpose):
+            return bool(int(self) < int(other))
+        if other.__class__ is int or other.__class__ is float:
+            return bool(int(self) < other)
+        return NotImplemented
+
+    def __ge__(self, other: Any) -> bool:
+        if (self.__class__ is other.__class__
+            or other.__class__ is Gtk.InputPurpose
+            or other.__class__ is IBus.InputPurpose):
+            return bool(int(self) >= int(other))
+        if other.__class__ is int or other.__class__ is float:
+            return bool(int(self) >= other)
+        return NotImplemented
+
+    def __le__(self, other: Any) -> bool:
+        if (self.__class__ is other.__class__
+            or other.__class__ is Gtk.InputPurpose
+            or other.__class__ is IBus.InputPurpose):
+            return bool(int(self) <= int(other))
+        if other.__class__ is int or other.__class__ is float:
+            return bool(int(self) <= other)
+        return NotImplemented
+
+    FREE_FORM = ('FREE_FORM')
+    ALPHA = ('ALPHA')
+    DIGITS = ('DIGITS')
+    NUMBER = ('NUMBER')
+    PHONE = ('PHONE')
+    URL = ('URL')
+    EMAIL = ('EMAIL')
+    NAME = ('NAME')
+    PASSWORD = ('PASSWORD')
+    PIN = ('PIN')
+    TERMINAL = ('TERMINAL')
+
+class InputHints(Flag):
+    '''Compatibility class to handle InputHints the same way no matter
+    what version of ibus is used.
+
+    For example, older versions of ibus might not have IBus.InputHints.PRIVATE
+    (or maybe even do not have IBus.InputHints at all). Then
+
+        input_hints & IBus.InputHints.PRIVATE
+
+    will produce an exception. But when using this compatibility class
+
+        input_hints & InputHints.PRIVATE
+
+    will just be False but not cause an exception.
+
+    See also:
+
+    https://docs.gtk.org/gtk3/flags.InputHints.html
+    https://docs.gtk.org/gtk4/flags.InputHints.html
+
+    Examples:
+
+    >>> int(InputHints.SPELLCHECK)
+    1
+
+    >>> InputHints.SPELLCHECK == 1
+    True
+
+    >>> InputHints.SPELLCHECK | 2
+    3
+
+    >>> 2 | InputHints.SPELLCHECK
+    3
+
+    >>> int(InputHints.NO_SPELLCHECK | InputHints.SPELLCHECK)
+    3
+
+    >>> 3 == InputHints.NO_SPELLCHECK | InputHints.SPELLCHECK
+    True
+
+    >>> 3 == InputHints.NO_SPELLCHECK | Gtk.InputHints.SPELLCHECK
+    True
+
+    >>> 3 == InputHints.NO_SPELLCHECK | IBus.InputHints.SPELLCHECK
+    True
+
+    >>> InputHints.SPELLCHECK == IBus.InputHints.SPELLCHECK
+    True
+
+    >>> InputHints.SPELLCHECK == Gtk.InputHints.SPELLCHECK
+    True
+    '''
+    def __new__(cls, attr: str) -> Any:
+        obj = object.__new__(cls)
+        if hasattr(Gtk, 'InputHints') and hasattr(Gtk.InputHints, attr):
+            obj._value_ = int(getattr(Gtk.InputHints, attr))
+        else:
+            obj._value_ = 0
+        return obj
+
+    def __int__(self) -> int:
+        return int(self._value_)
+
+    def __eq__(self, other: Any) -> bool:
+        if (self.__class__ is other.__class__
+            or other.__class__ is Gtk.InputHints
+            or other.__class__ is IBus.InputHints):
+            return bool(int(self) == int(other))
+        if other.__class__ is int or other.__class__ is float:
+            return bool(int(self) == other)
+        return NotImplemented
+
+    def __or__(self, other: Any) -> Any:
+        if self.__class__ is other.__class__:
+            return self.value | other.value
+        if (other.__class__ is Gtk.InputHints
+            or other.__class__ is IBus.InputHints):
+            return int(self) | int(other)
+        if other.__class__ is int:
+            return int(self) | other
+        return NotImplemented
+
+    def __ror__(self, other: Any) -> Any: # type: ignore[override]
+        return self.__or__(other)
+
+    def __and__(self, other: Any) -> Any:
+        if self.__class__ is other.__class__:
+            return self.value & other.value
+        if (other.__class__ is Gtk.InputHints
+            or other.__class__ is IBus.InputHints):
+            return int(self) & int(other)
+        if other.__class__ is int:
+            return int(self) & other
+        return NotImplemented
+
+    def __rand__(self, other: Any) -> Any: # type: ignore[override]
+        return self.__and__(other)
+
+    NONE = ('NONE')
+    SPELLCHECK = ('SPELLCHECK')
+    NO_SPELLCHECK = ('NO_SPELLCHECK')
+    WORD_COMPLETION = ('WORD_COMPLETION')
+    LOWERCASE = ('LOWERCASE')
+    UPPERCASE_CHARS = ('UPPERCASE_CHARS')
+    UPPERCASE_WORDS = ('UPPERCASE_WORDS')
+    UPPERCASE_SENTENCES = ('UPPERCASE_SENTENCES')
+    INHIBIT_OSK = ('INHIBIT_OSK')
+    VERTICAL_WRITING = ('VERTICAL_WRITING')
+    EMOJI = ('EMOJI')
+    NO_EMOJI = ('NO_EMOJI')
+    PRIVATE = ('PRIVATE')
+
 class KeyEvent:
     '''Key event class used to make the checking of details of the key
     event easy
@@ -353,6 +667,7 @@ class KeyEvent:
         self.mod1 = self.state & IBus.ModifierType.MOD1_MASK != 0
         # mod2: Usually Num_Lock (0x4d)
         self.mod2 = self.state & IBus.ModifierType.MOD2_MASK != 0
+        # mod3: Usually Scroll_Lock
         self.mod3 = self.state & IBus.ModifierType.MOD3_MASK != 0
         # mod4: Usually Super_L (0xce),  Hyper_L (0xcf)
         self.mod4 = self.state & IBus.ModifierType.MOD4_MASK != 0
@@ -386,24 +701,30 @@ class KeyEvent:
         return False
 
     def __str__(self) -> str:
-        return (
-            "val=%s code=%s state=0x%08x name='%s' unicode='%s' "
-            % (self.val,
-               self.code,
-               self.state,
-               self.name,
-               self.unicode)
-            + "shift=%s lock=%s control=%s super=%s hyper=%s meta=%s "
-            % (self.shift,
-               self.lock,
-               self.control,
-               self.super,
-               self.hyper,
-               self.meta)
-            + "mod1=%s mod5=%s release=%s\n"
-            % (self.mod1,
-               self.mod5,
-               self.release))
+        return repr(
+            f'val={self.val} '
+            f'code={self.code} '
+            f'state=0x{self.state:08x} '
+            f'name=“{self.name}” '
+            f'unicode=“{self.unicode}” '
+            f'shift={self.shift} '
+            f'lock={self.lock} '
+            f'control={self.control} '
+            f'super={self.super} '
+            f'hyper={self.hyper} '
+            f'meta={self.meta} '
+            f'mod1={self.mod1} '
+            f'mod2={self.mod2} '
+            f'mod3={self.mod3} '
+            f'mod4={self.mod4} '
+            f'mod5={self.mod5} '
+            f'button1={self.button1} '
+            f'button2={self.button2} '
+            f'button3={self.button3} '
+            f'button4={self.button4} '
+            f'button5={self.button5} '
+            f'release={self.release} '
+            f'modifier={self.modifier}')
 
 def keyevent_to_keybinding(keyevent: KeyEvent) -> str:
     keybinding = ''
@@ -530,14 +851,14 @@ class HotKeys:
     def __str__(self) -> str:
         return repr(self._hotkeys)
 
-class ItKeyInputDialog(Gtk.MessageDialog):
+class ItKeyInputDialog(Gtk.MessageDialog): # type: ignore
     def __init__(
             self,
             # Translators: This is used in the title bar of a dialog window
             # requesting that the user types a key to be used as a new
             # key binding for a command.
-            title=_('Key input'),
-            parent=None) -> None:
+            title: str = _('Key input'),
+            parent: Gtk.Window = None) -> None:
         Gtk.Dialog.__init__(
             self,
             title=title,
@@ -562,18 +883,18 @@ class ItKeyInputDialog(Gtk.MessageDialog):
         self.show()
 
     def on_key_press_event(# pylint: disable=no-self-use
-            self, widget, event) -> bool:
+            self, widget: Gtk.MessageDialog, event: Gdk.EventKey) -> bool:
         widget.e = (event.keyval,
                     event.get_state() & KEYBINDING_STATE_MASK)
         return True
 
     def on_key_release_event(# pylint: disable=no-self-use
-            self, widget, _event) -> bool:
+            self, widget: Gtk.MessageDialog, _event: Gdk.EventKey) -> bool:
         widget.response(Gtk.ResponseType.OK)
         return True
 
-class ItAboutDialog(Gtk.AboutDialog):
-    def  __init__(self, parent=None) -> None:
+class ItAboutDialog(Gtk.AboutDialog): # type: ignore
+    def  __init__(self, parent: Gtk.Window = None) -> None:
         Gtk.AboutDialog.__init__(self, parent=parent)
         self.set_modal(True)
         # An empty string in aboutdialog.set_logo_icon_name('')
@@ -589,7 +910,7 @@ class ItAboutDialog(Gtk.AboutDialog):
             _('Table input method for IBus.'))
         self.set_copyright(
             'Copyright © 2009-2012 Peng Huang,\n'
-            'Copyright © 2012-2020 Mike FABIAN')
+            'Copyright © 2012-2022 Mike FABIAN')
         self.set_authors([
             'Yuwei YU (“acevery”)',
             'Peng Huang',
@@ -643,14 +964,12 @@ class ItAboutDialog(Gtk.AboutDialog):
         self.show()
 
     def on_close_aboutdialog( # pylint: disable=no-self-use
-            self, _about_dialog, _response) -> None:
+            self, _about_dialog: Gtk.Dialog, _response: Gtk.ResponseType) -> None:
         '''
         The “About” dialog has been closed by the user
 
         :param _about_dialog: The “About” dialog
-        :type _about_dialog: GtkDialog object
         :param _response: The response when the “About” dialog was closed
-        :type _response: Gtk.ResponseType enum
         '''
         self.destroy()
 
